/*
** Module   : AgentPool.java
** Abstract : Definition and implementation of an appserver agent.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130628 Fixed context-local data reset for State-reset operating mode.
** 003 CA  20130701 Although official docs mention that State-reset doesn't call the shutdown or
**                  startup procedure, testing showed startup/shutdown procedures are called for  
**                  this operating mode too.
**                  Reworked appserver launching - implemented by the AppServerLauncher class.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.Array;
import java.util.*;

import org.apache.commons.logging.*;
import com.goldencode.p2j.NumberedException;
import com.goldencode.p2j.main.AppServerInvocationResult;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.RestrictedUseException;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.BlockManager.Condition;

/**
 * An apserver agent definition.
 */
class Agent
{
   /**
    * List of possible invocation modes: 
    * <ul>
    * <li>{@link #IMMEDIATELY} - in this thread (the thread must be an agent thread)</li>
    * <li>{@link #AGENT_WAIT} - in a different agent, wait for response</li>
    * <li>{@link #AGENT_NO_WAIT} in a different agent, do not wait for response</li>
    * </ul>
    */
   enum InvocationModes
   {
      IMMEDIATELY,
      AGENT_WAIT,
      AGENT_NO_WAIT
   }
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerManager.class);

   /** The command queue for this agent. */
   private final Queue<AgentCommand> commands = new LinkedList<AgentCommand>();
   
   /** The agent's pool. */
   private final AgentPool pool;

   /** This agent's thread. */
   private final Thread thread = Thread.currentThread();

   /** Flag indicating if the agent has been terminated. */
   private boolean terminated = false;
   
   /** Flag indicating the agent has been bound to the connection. */
   private boolean bound = false;
   
   /**
    * Build a new agent and associate it with the given pool.
    * 
    * @param    pool
    *           The agent's pool.
    * @param    propath
    *           The appserver's default propath.
    */
   public Agent(AgentPool pool, String propath)
   {
      this.pool = pool;
      // set the propath
      EnvironmentOps.setSearchPath(propath);
   }

   /**
    * Invoke an external procedure with no parameters. 
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    */
   private static void invokeScoped(String id, character name)
   {
      invokeScoped(id, name, null, false, false, false, false, null, null);
   }
   
   /**
    * Invoke an external procedure, internal procedure or user-defined function.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    proc
    *           The external procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    exports
    *           The appserver's exports.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    The result returned by this procedure or user-defined function call. 
    */
   private static BaseDataType invokeScoped(String    id,
                                            character name,
                                            handle    proc,
                                            boolean   function,
                                            boolean   dynamicFunction,
                                            boolean   superCall,
                                            boolean   transactionDistinct,
                                            character exports,
                                            String    modes,
                                            Object... args)
   {
      // add the topmost scope to the TransactionManager
      TransactionManager.pushScope("startup",
                                   TransactionManager.NO_TRANSACTION,
                                   true,
                                   true,
                                   false,
                                   false);
      try
      {
         if (proc != null)
         {
            ProcedureManager.pushThisProcedure(proc);
         }
   
         if (proc == null)
         {
            // force invocation of an external procedure
            ControlFlowOps.invokeExternalProcedure(name, false, null, transactionDistinct,
                                                   exports, modes, args);
            return ControlFlowOps.getReturnValue();
         }
         else
         {
            // this is an internal-entry call
            return ControlFlowOps.invoke(proc, name, function,
                                         dynamicFunction, superCall,
                                         transactionDistinct, modes, args);
         }
      }
      catch (StopConditionException e)
      {
         // check if this can be converted to error
         if (exports != null  && invokeFailure(e))
         {
            return new unknown();
         }
         
         throw e;
      }
      finally
      {
         if (proc != null)
         {
            ProcedureManager.popThisProcedure();
         }
         // remove the topmost scope from the TransactionManager
         TransactionManager.popScope();
      }
   }

   /**
    * Check if the given STOP condition is from an invoke failure.
    * 
    * @param    e
    *           The STOP condition.
    * 
    * @return   <code>true</code> if the cause if this STOP is a {@link NumberedException} known
    *           to be produced by invoke failures.
    */
   private static boolean invokeFailure(StopConditionException e)
   {
      Throwable cause = e.getCause();
      if (cause != null && cause instanceof NumberedException)
      {
         // this specific case treats the error raised by ControlFlowOps.invokeFailure, which will
         // always set the topmost cause as a NumberedException
         NumberedException ne = (NumberedException) cause;
         if (ne.getNumber() == 293)
         {
            // raise ERROR condition if the external procedure was not found
            throw new ErrorConditionException(ne);
         }
      }
      
      return false;
   }

   /**
    * Start the agent and listen for commands.
    */
   public void listen()
   {
      // register a listener to terminate the agent, if its session terminates
      Session sess = SessionManager.get().getSession();
      sess.addSessionListener(new SessionListener()
      {
         @Override
         public void terminate(Session session)
         {
            Agent.this.terminate();
         }
      });
      
      prepare();

      AppServerOperatingMode opMode = pool.getOperatingMode();
      String startup = pool.getStartup();
      // although official docs mention that state-reset doesn't call the startup procedure, 
      // testing showed that startup procedure is called
      handle hStartup = new handle();
      if (startup != null && startup.length() > 0)
      {
         String startupParameter = pool.getStartupParameter();
         if (!invoke(null, null, InvocationModes.IMMEDIATELY, startup, hStartup, "Startup",
                     "I", startupParameter))
         {
            // if there were problems invoking the startup procedure, terminate the agent
            return;
         }
      }
   
      // notify the launch
      AppServerLauncher.notifyLaunch(pool.getName(), true);
      
      while (!terminated())
      {
         AgentCommand command = null;
         
         // this loop will continue until someone sends the terminate command:
         // - in case the session is ended (P2J client gets killed)
         // - in case the manager is ended
         
         boolean interrupted = false;
         synchronized (this)
         {
            if (commands.size() == 0)
            {
               try
               {
                  wait();
               }
               catch (InterruptedException e)
               {
                  interrupted = true;
                  if (LOG.isWarnEnabled())
                  {
                     LOG.warn("Agent was interrupted for appserver " + pool.getName(), e);
                  }
               }
            }
   
            // get the last command (or null if none)
            command = commands.poll();
         }

         // someone has interrupted my waiting, check if the session is still alive
         if (interrupted && !sess.isRunning())
            return;
         
         if (command == null)
         {
            // no command yet, continue
            continue;
         }
         
         try
         {
            command.execute();
         }
         catch (Throwable t)
         {
            if (LOG.isWarnEnabled())
            {
               LOG.warn("Agent encountered an error while executing a command for appserver " + 
                        pool.getName(), t);
            }
            command.setError(t);
         }
         finally
         {
            command.finished();
         }
      }
      
      String shutdown = pool.getShutdown();
      // although official docs mention that state-reset doesn't call the shutdown procedure, 
      // testing showed that shutdown procedure is called
      if (shutdown != null && shutdown.length() > 0)
      {
         // ignore errors here
         invoke(null, null, InvocationModes.IMMEDIATELY, shutdown, null, "Shutdown", null);
      }
   
      // delete the startup procedure handle
      if (hStartup._isValid())
      {
         HandleOps.delete(hStartup);
      }
   }

   /**
    * Reset the context of this agent.  Called only if the appserver is in State-reset operating
    * mode.
    */
   public void resetContext()
   {
      waitForCommand(null, new ResetContextCommand());
   }
   
   /**
    * Establish a new connection and call the CONNECT procedure.
    * 
    * @return   The connection ID.
    * 
    * @throws   NumberedException
    *           If a connection could not be established.
    */
   public String connect()
   throws NumberedException
   {
      final character id = new character();
      final AppServerInvocationResult result = new AppServerInvocationResult();
      waitForCommand(null, new AgentCommand()
      {
         @Override
         public void execute()
         {
            // establish a connection
            try
            {
               id.assign(pool.newConnection());
            }
            catch (NumberedException e)
            {
               result.setError(e);
            }
         }
      });

      Throwable t = result.getError();
      if (t != null)
      {
         if (t instanceof NumberedException)
         {
            throw (NumberedException) t;
         }
         else
         {
            id.setUnknown();
         }
      }
      
      return id.isUnknown() ? null : id.toStringMessage();
   }

   /**
    * Invoke the specified external procedure as persistent and wait (or not) for it to complete.
    * This is used to invoke any connect/disconnect and startup/shutdown procedures.
    * <p>
    * The "kind" parameter will always receive one of the "Startup", "Shutdown", "Connect" or 
    * "Disconnect" values.
    * 
    * @param    res
    *           The instance where to save the invocation result. May be <code>null</code>.
    * @param    id
    *           The appserver connection id. 
    * @param    mode
    *           Flag indicating how to invoke the external procedure.
    * @param    proc
    *           The legacy 4GL name for the external procedure.
    * @param    hProc
    *           If not-null, the handle where to save this procedure.
    * @param    kind
    *           A helper text about this procedure.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   <code>true</code> if no error was raised during the procedure invocation.
    */
   public boolean invoke(AppServerInvocationResult res,
                         final String              id,
                         final InvocationModes     mode,
                         final String              proc,
                         final handle              hProc,
                         final String              kind,
                         final String              modes,
                         final Object...           args)
   {
      final AppServerInvocationResult result = (res == null ? new AppServerInvocationResult() 
                                                            : res);

      AgentCommand command = new AgentCommand() 
      {
         @Override
         public void execute()
         {
            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            Throwable error = null;
            try
            {
               ControlFlowOps.invokePersistentSetWithMode(proc, hProc, modes, args);
            }
            catch (Throwable t)
            {
               error = t;
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
   
            if (error != null && LOG.isWarnEnabled())
            {
               // procedure has failed
               LOG.warn(String.format("%s procedure has failed for appserver %s:",
                                      kind, pool.getName()), 
                        error);
            }
            
            if (result != null)
            {
               result.setError(error);
               result.setResult(ControlFlowOps.getReturnValue());
            }
            
            // check if we are calling the disconnect procedure
            if ("Disconnect".equals(kind))
            {
               pool.terminateConnection(id);
            }
         }
      };
      
      switch (mode)
      {
         case IMMEDIATELY:
            try
            {
               command.execute();
            }
            catch (Throwable t)
            {
               result.setError(t);
            }
            break;

         case AGENT_NO_WAIT:
            sendCommand(command);
            break;

         case AGENT_WAIT:
            waitForCommand(result, command);
            break;
      }
      
      return (result.getError() == null);
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    resId
    *           The code of a remote procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invoke(final String    id,
                                           final character name,
                                           final String    resId,
                                           final boolean   function,
                                           final boolean   dynamicFunction,
                                           final boolean   superCall,
                                           final boolean   transactionDistinct,
                                           final String    modes,
                                           final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand()
      {
         @Override
         public void execute()
         {
            String activate = pool.getActivate();
            if (!isBound() && activate != null && activate.length() > 0)
            {
               // call the activate procedure (only if the agent is not already bound)
               try
               {
                  invokeScoped(id, new character(activate));
               }
               catch (ConditionException ce)
               {
                  Condition c = BlockManager.exceptionToCondition(ce);
                  String type = c.toString();
                  int code = 0;
                  switch (c)
                  {
                     case ERROR:
                        code = 8025;
                        break;
                     case STOP:
                        code = 8026;
                        break;
                     case QUIT:
                        code = 8027;
                        break;
                  }

                  final String msg = "%s  activation procedure ended with an %s condition";
                  invocationResult.setError(
                     new NumberedException(String.format(msg, activate, type), code));
               }
               catch (Throwable t)
               {
                  invocationResult.setError(t);
               }
               finally
               {
                  // pass return value
                  invocationResult.setResult(ControlFlowOps.getReturnValue());
                  
                  // if an error was encountered during activate, call deactivate and disconnect
                  if (invocationResult.getError() != null)
                  {
                     String deactivate = pool.getDeactivate();
                     if (!isBound() && deactivate != null && deactivate.length() > 0)
                     {
                        // call the deactivate procedure (only if the agent is not bound)
                        try
                        {
                           invokeScoped(id, new character(deactivate));
                        }
                        catch (Throwable t)
                        {
                           // ignore all errors
                        }
                     }

                     AppServerManager.disconnect(id);
                     
                     return;
                  }
               }
            }
            
            // resolve the procedure
            handle proc = (resId == null ? null : new handle(pool.getProcedure(id, resId)));
            character exports = pool.getExports();
            boolean disconnect = false;
            String localModes = null;
            if (args.length > 0)
            {
               if (proc == null)
               {
                  localModes = SourceNameMapper.getParameterModes(name.toStringMessage());
               }
               else if (ProcedureManager.isProcedure(proc))
               {
                  String pname = handle.unwrapProcedure(proc).getName().toStringMessage();
                  String iename = name.toStringMessage();
                  localModes = SourceNameMapper.getParameterModes(pname, iename, function);
               }
            }

            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            try
            {
               preProcessArguments(localModes, modes, args);

               Object result = invokeScoped(id, name, proc,
                                            function, dynamicFunction, superCall,
                                            transactionDistinct, exports, modes, args);
               if (!function)
               {
                  result = ControlFlowOps.getReturnValue();
               }
               
               if (result instanceof handle)
               {
                  // handle results are converted to a container holding only the remote resource
                  handle h = (handle) result;
                  result = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
               }
               invocationResult.setResult(result);
            }
            catch (ErrorConditionException e)
            {
               Throwable cause = e.getCause();
               if (cause instanceof NumberedException)
               {
                  // only the following codes will be sent to the remote side
                  // TODO: this needs to be done only if the errors are caused by problems in
                  // resolving the appserver call target, and not by nested RUN statements
                  final Set<Integer> notifyErrs = new HashSet<Integer>();
                  final Integer[] codes = 
                     { 6349, 3230, 1005, 3234, 5729, 2570, 11428, 12290, 15472, 5639, 6445 };
                  notifyErrs.addAll(Arrays.asList(codes));

                  NumberedException ne = (NumberedException) cause;
                  int num = ne.getNumber();
                  if (notifyErrs.contains(num))
                  {
                     invocationResult.setError(e);
                  }
               }
            }
            catch (QuitConditionException e)
            {
               disconnect = true;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(localModes, modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);

               String deactivate = pool.getDeactivate();
               if (!isBound() && deactivate != null && deactivate.length() > 0)
               {
                  // call the deactivate procedure (only if the agent is not bound)
                  try
                  {
                     invokeScoped(id, new character(deactivate));
                  }
                  catch (Throwable t)
                  {
                     // ignore all errors
                  }
               }

               if (disconnect)
               {
                  AppServerManager.disconnect(id);
               }
            }
         }
      });
      
      return invocationResult;
   }

   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *
    * @return    A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *            procedure, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    */
   public AppServerInvocationResult invokePersistent(final String    id,
                                                     final character name,
                                                     final boolean   transactionDistinct,
                                                     final String    modes,
                                                     final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand()
      {
         @Override
         public void execute()
         {
            final handle hProc = new handle();

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            String localModes = null;
            if (args.length > 0)
            {
               localModes = SourceNameMapper.getParameterModes(name.toStringMessage());
            }

            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            try
            {
               preProcessArguments(localModes, modes, args);
               
               character exports = pool.getExports();
               ControlFlowOps.invokeExternalProcedure(name, true, hProc, transactionDistinct,
                                                      exports, modes, args);
            }
            catch (StopConditionException e)
            {
               if (invokeFailure(e))
               {
                  return;
               }
               
               throw e;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(localModes, modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);

               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }

            // save this handle globally, as it needs to be exposed to the conversation thread
            // which handles the AppServerEntry calls too
            if (id != null)
            {
               pool.saveProcedure(id,
                                  hProc.toString(),
                                  (PersistentProcedure) hProc.getResource());
            }

            // make it bound only if the invocation was successful
            bind();
            
            invocationResult.setResult(ControlFlowOps.getReturnValue());
            invocationResult.setProcedureId(hProc.toString());
         }
      });

      return invocationResult;
   }

   /**
    * Send the terminate command to this agent.
    */
   public void terminate()
   {
      // interrupt current work
      thread.interrupt();
      
      // send the terminate command
      sendCommand(new AgentCommand()
      {
         @Override
         public void execute()
         {
            synchronized (Agent.this)
            {
               terminated = true;
            }
         }
      });
   }

   /**
    * Check if this agent has been terminated.
    * 
    * @return   See above.
    */
   public synchronized boolean terminated()
   {
      return terminated;
   }
   
   /**
    * Bind the agent to the connection.
    */
   public synchronized void bind()
   {
      this.bound = true;
   }
   
   /**
    * Unbind the agent from the connection.
    */
   public synchronized void unbind()
   {
      this.bound = false;
   }
   
   /**
    * Check if the agent is bound to the connection.
    * 
    * @return   See above.
    */
   public synchronized boolean isBound()
   {
      return bound;
   }
   
   /**
    * Send the given command to this agent and return immediately.
    * 
    * @param    command
    *           The command to be executed.
    */
   private synchronized void sendCommand(AgentCommand command)
   {
      commands.add(command);
      
      this.notify();
   }
   
   /**
    * Send the given command to the agent and wait for it to complete.
    * 
    * @param    result
    *           The instance where to place the command result.
    * @param    command
    *           The command to be executed.
    */
   private void waitForCommand(AppServerInvocationResult result, AgentCommand command)
   {
      sendCommand(command);

      while (!command.isFinished())
      {
         synchronized (command)
         {
            try
            {
               command.wait();
            }
            catch (InterruptedException e)
            {
               if (LOG.isWarnEnabled())
               {
                  LOG.warn("Command was interrupted for appserver " + pool.getName(), e);
               }
            }
         }
      }

      if (result != null && command.getError() != null)
      {
         result.setError(command.getError());
      }
   }
   
   /**
    * Preprocess the arguments sent by the requester so that they are converted to instances
    * compatible with the P2J runtime.  This includes:
    * <ul>
    * <li>Creating the handles for all received resources</li>
    * <li>Creating the temp-tables for the received {@link TableWrapper} instances</li>
    * <li>Creating {@link memptr} instances from all {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    localModes
    *           The parameter modes as defined at the external procedure or internal entry.
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    args
    *           The argument list sent by the requester.  The elements in this array will be
    *           changed as needed.
    */
   private void preProcessArguments(String localModes, String requesterModes, Object[] args)
   {
      if (localModes != null && requesterModes != null && !localModes.equals(requesterModes))
      {
         // invocation will fail, so nothing to prepare
         return;
      }
      
      String modes = (localModes == null ? requesterModes : localModes);
      if (modes == null || modes.length() != args.length)
      {
         // no modes specified or incompatible modes
         return;
      }
      
      for (int i = 0; i < modes.length(); i++)
      {
         Object arg = args[i];

         if (arg instanceof TableWrapper)
         {
            // a result set will be sent only if the requester and the appserver are not in the  
            // same JVM
            TableWrapper rs = (TableWrapper) arg;

            Temporary dmo = null;
            // TODO: create the backing temp-table as a dynamic temp-table, and populate its data

            arg = new TableParameter(dmo, rs.isAppend());
         }
         else
         {
            boolean isArray = arg.getClass().isArray();
            Object[] data = null;
            
            if (isArray)
            {
               // check if this argument is an array
               data = new Object[Array.getLength(arg)];
               
               for (int j = 0; j < data.length; j++)
               {
                  data[j] = Array.get(arg, j);
               }
            }
            else
            {
               data = new Object[] { arg };
            }
            
            for (int j = 0; j < data.length; j++)
            {
               Object darg = data[j];
               
               if (darg instanceof RemoteResource)
               {
                  String resid = ((RemoteResource) darg).getResourceId();
                  darg = (resid == null ? new handle() : handle.fromString(resid));
               }
               else if (darg instanceof MemoryBuffer)
               {
                  // convert MemoryBuffer instances to memptr instances
                  darg = new memptr(((MemoryBuffer) darg).getValue());
               }
               
               data[j] = darg;
            }
            
            arg = (isArray ? data : data[0]);
         }
         
         args[i] = arg;
      }
   }

   /**
    * Post-process the given arguments so that all OUTPUT and INPUT-OUTPUT arguments will be sent
    * back to the requester. This includes:
    * <ul>
    *    <li>Serializing all table parmaeters via {@link TableWrapper} instances</li>
    *    <li>Sending the resource IDs for all handle parameters</li>
    *    <li>Converting all {@link memptr} instances to {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    localModes
    *           The parameter modes as defined at the external procedure or internal entry.
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    localArgs
    *           The argument list after the call has been performed.
    * @param    requesterArgs
    *           The argument list sent by the requester.
    * 
    * @return   The postprocessed arguments, to be sent back to the requester.
    */
   private Object[] postProcessArguments(String   localModes,
                                         String   requesterModes,
                                         Object[] localArgs,
                                         Object[] requesterArgs)
   {
      if (localModes != null && requesterModes != null && !localModes.equals(requesterModes))
      {
         // invocation will fail, so nothing to prepare
         return new Object[0];
      }
      
      String modes = (localModes == null ? requesterModes : localModes);
      if (modes == null || requesterArgs.length != modes.length())
      {
         // no modes specified, so no arguments can be returned - assume all are input
         return new Object[0];
      }
      
      Object[] arguments = new Object[localArgs.length];
      for (int i = 0; i < modes.length(); i++)
      {
         char mode = modes.charAt(i);

         // in input mode, the argument is not sent back
         Object localArg = (mode == 'O' || mode == 'U' ? localArgs[i] : null);
         
         if (localArg == null)
         {
            // nothing to post-process if the local arg was determined to be null
            arguments[i] = null;
            continue;
         }
         Object requesterArg = requesterArgs[i];

         // if output/input-output mode, send back the data. 

         // a table parameter is serialized to a result set only if the requester has sent a 
         // result set
         if (localArg instanceof TableParameter && requesterArg instanceof TableWrapper)
         {
            // we have a DMO on the remote side, use the P2J persistence layer to get the data
            TableParameter tableParam = (TableParameter) localArg;
            Temporary tempDMO = tableParam.getTable();
            boolean append = tableParam.isAppend();
            boolean input = (mode == 'U');

            try
            {
               localArg = new TableWrapper(new TempTableResultSet(tempDMO, input, true, append));
            }
            catch (ClassNotFoundException e)
            {
               // should never happen, ignore it but log it anyway
               if (LOG.isErrorEnabled())
               {
                  LOG.error("Error building a TempTableResultSet for parameter on position " + i,
                            e); 
               }
            }
         }
         else
         {
            boolean isArray = localArg.getClass().isArray();
            Object[] data = null;
            
            if (isArray)
            {
               // check if this argument is an array
               data = new Object[Array.getLength(localArg)];
               
               for (int j = 0; j < data.length; j++)
               {
                  data[j] = Array.get(localArg, j);
               }
            }
            else
            {
               data = new Object[] { localArg };
            }
            
            for (int j = 0; j < data.length; j++)
            {
               Object dlocalArg = data[j];
               
               if (dlocalArg instanceof handle)
               {
                  handle h = (handle) dlocalArg;
                  
                  dlocalArg = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
               }
               else if (dlocalArg instanceof memptr)
               {
                  // convert memptr instances to MemoryBuffer instances
                  dlocalArg = new MemoryBuffer(((memptr) dlocalArg).asByteArray());
               }
               
               data[j] = dlocalArg;
            }
            
            localArg = (isArray ? data : data[0]);
         }
         
         arguments[i] = localArg;
      }
      
      return arguments;
   }

   /**
    * Prepare this agent by setting some batch-processing related configuration.
    */
   private void prepare()
   {
      // avoid the end-of-program message
      LogicalTerminal.setRelogin(true);
      
      // set this as batch mode
      EnvironmentOps.setBatchMode(true);
   }
   
   /**
    * Implements the reset context command.
    */
   private class ResetContextCommand
   extends AgentCommand
   {
      @Override
      public void execute()
      {
         // reset the context
         try
         {
            SecurityManager.getInstance().resetContext();
         }
         catch (RestrictedUseException e)
         {
            // ignore
         }
         
         // set some state again
         prepare();
      }
   }
   
   /**
    * Defins a command to be sent to an agent.
    */
   private static abstract class AgentCommand
   {
      /** Any error caught during the command execution. */
      private Throwable error = null;
      
      /** Flag indicating if this command has finished or not. */
      private boolean finished = false;
      
      /** Override this to define the command body. */
      public abstract void execute();
      
      /**
       * Check the {@link #finished} flag.
       * 
       * @return   See above.
       */
      public synchronized boolean isFinished()
      {
         return finished;
      }

      /**
       * Notify that this command has finished.
       */
      public synchronized void finished()
      {
         this.finished = true;
         this.notify();
      }
      
      /**
       * Set the error caught during this command execution.
       * 
       * @param    t
       *           The caught error.
       */
      public synchronized void setError(Throwable t)
      {
         this.error = t;
      }
      
      /**
       * Get the error caught during this command execution.
       * 
       * @return   See above.
       */
      public synchronized Throwable getError()
      {
         return this.error;
      }
   }
}
