/*
** Module   : AppServerLauncher.java
** Abstract : Class dedicated to launching appservers.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130705 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.management.*;
import java.net.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * This is a launcher of appservers and is used in two ways:
 * <ul>
 *    <li>Automatically starts the appservers which have the <code>auto_start=true</code> in their
 *        configuration; this will start up to <code>initial_agents</code>.</li>
 *    <li>Allow starting of appservers remotely; if a server is started using the "-a appserver"
 *        argument, it will connect to the target P2J Server and call {@link AppServer#launch} to
 *        launch the specified appserver. If the appserver is already started, it will attempt to
 *        start an agent.</li>
 * </ul>
 * 
 * There is one other way to start an appserver: by manually specifying the ID of an account 
 * configured to start an appserver, when starting a P2J Client.  This will start only one Agent
 * (if non started).
 */
public final class AppServerLauncher
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerLauncher.class);

   /** The maximum number of continous failures to start an Agent, before giving up. */
   private static final int MAX_FAILURES = 10;
   
   /** The maximum amount of time to wait for an Agent to start. */
   private static final int LAUNCH_TIMEOUT = 120000;
   
   /** Map of launchers, per appserver name. */
   private static final Map<String, Launcher> launchers = new HashMap<String, Launcher>();
   
   /**
    * Collect all the appservers configured with <code>auto_start=true</code> and attempt to start
    * them.  The Agent launching will be done in a distinct thread, with no associated context.
    * <p>
    * Note that an appserver which has no account configured for it or has more than one account
    * configured for it will fail to start.
    */
   public static void startAppservers()
   {
      final Map<String, String> appServers = new HashMap<String, String>();
      SecurityManager secMgr = SecurityManager.getInstance();

      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
         throw new RuntimeException("directory bind failed");

      // read the appservers
      String appSrvPath = Utils.findDirectoryNodePath(ds, "appservers", "container", false);
      String[] appservers = ds.enumerateNodes(appSrvPath);
      for (int i = 0; i < appservers.length; i++)
      {
         String appServer = appservers[i];
         
         // find the user of this appserver
         String[] subjects = secMgr.getAccountsForAppserver(appServer);
         if (subjects.length == 0)
         {
            if (LOG.isWarnEnabled())
            {
               final String msg = "Appserver '%s' has no associated user.";
               LOG.warn(String.format(msg, appServer));
            }

            continue;
         }
         else if (subjects.length > 1)
         {
            if (LOG.isWarnEnabled())
            {
               final String msg = "Appserver '%s' is associated with more than one user: %s";
               LOG.warn(String.format(msg, appServer, Arrays.asList(subjects)));
            }
            
            continue;
         }
         
         // appserver is configured OK
         Boolean autoStart = ds.getNodeBoolean(appSrvPath + "/" + appServer, "auto_start");
         if (autoStart == null || !autoStart)
         {
            continue;
         }
         
         // configured for auto start, record it
         appServers.put(appServer, subjects[0]);
         
         // force the loading of the appserver definition in this thread, as the launcher thread
         // has no context set
         AppServerDefinition.get(appServer);
      }

      ds.unbind();

      Thread t = new Thread()
      {
         @Override
         public void run()
         {
            for (String appserver : appServers.keySet())
            {
               String account = appServers.get(appserver);

               launch(account, appserver);
            }
         }
      };

      t.setName("AppServer Launcher");
      t.setDaemon(true);
      t.start();
   }
   
   /**
    * Launch the specified appserver. This includes starting the specified number of initial 
    * agents.
    * 
    * @param    appServer
    *           The appserver name to launch.
    * 
    * @return   <code>true</code> if the appserver was launched.
    */
   public static boolean launch(String appServer)
   {
      SecurityManager secMgr = SecurityManager.getInstance();
   
      // find the user of this appserver
      String[] subjects = secMgr.getAccountsForAppserver(appServer);
      if (subjects.length == 0)
      {
         final String msg = "Appserver '%s' has no associated user.";
         throw new IllegalStateException(String.format(msg, appServer));
      }
      else if (subjects.length > 1)
      {
         final String msg = "Appserver '%s' is associated with more than one user: %s";
         throw new IllegalStateException(String.format(msg, 
                                                       appServer,
                                                       Arrays.asList(subjects)));
      }
      
      return launch(subjects[0], appServer);
   }

   /**
    * When the appserver Agent is actually about to start listening for commands, it will notify 
    * its launcher that it was started. Until then, the appserver Agent is presumed "not running".
    * <p>
    * Note that when an appserver is configured with <code>auto_start=false</code> and the agent
    * is started manually, no {@link Launcher} will be available.  Launchers are used only the 
    * appserver is started automatically or when the server receives the "-p appserver" argument.
    * 
    * @param    appServer
    *           The name of the appserver with which the Agent is associated.
    * @param    success
    *           <code>true</code> if the Agent was started.
    */
   static void notifyLaunch(String appServer, boolean success)
   {
      Launcher launcher = null;
      synchronized (launchers)
      {
         launcher = launchers.get(appServer);
      }
      
      // the launcher can be null if started manually
      if (launcher != null)
      {
         synchronized (launcher)
         {
            launcher.success = success;
   
            launcher.notify();
         }
      }
   }
   
   /**
    * Launch an agent for the specified appserver.  The agent will need to authenticate using the
    * specified account.
    * <p>
    * If no agent is alive for this appserver, it will initialize the appserver by starting an
    * <code>initial_agents</code> Agents.  An Agent startup will fail if it takes more than 
    * {@value #LAUNCH_TIMEOUT} seconds.  Also, the Agent startup will not continue if it failed
    * more than {@link #MAX_FAILURES} times in a row.
    * <p>
    * If first call and at least one agent has been started, it will start a thread responsible
    * of sending the 
    * 
    * @param    account
    *           The account used to authenticate the Agent.
    * @param    appServer
    *           The appserver name.
    *           
    * @return   <code>true</code> if the appserver or agent was started successfully.
    */
   private static boolean launch(String account, final String appServer)
   {
      Launcher launcher = null;
      boolean first = false;
      synchronized (launchers)
      {
         launcher = launchers.get(appServer);
         if (launcher == null)
         {
            first = true;
            launcher = new Launcher(builder(account, appServer));
            launchers.put(appServer, launcher);
         }
      }
      
      synchronized (launcher)
      {
         launcher.success = null;
      }
      
      AppServerDefinition appDef = AppServerDefinition.get(appServer);
      boolean success = true;
      if (first)
      {
         int failures = 0;
         int initialAgents = appDef.getInitialAgents();
         while (true)
         {
            boolean launched = launchWorker(appServer, launcher);
            failures = (launched ? 0 : failures + 1);
            
            AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
            if (pool != null && pool.size() >= initialAgents)
            {
               break;
            }
            
            if (failures >= MAX_FAILURES)
            {
               // protect this loop so it is finite; if agent startup failed 10 times in a row,
               // abend the process

               if (LOG.isWarnEnabled())
               {
                  if (pool == null)
                  {
                     final String msg = "Could not start appserver '%s'.";
                     LOG.warn(String.format(msg, appServer));
                     success = false;
                  }
                  else
                  {
                     final String msg = "Only %d out of %d were started for appserver %s";
                     LOG.warn(String.format(msg, pool.size(), initialAgents, appServer));
                  }
               }
               
               break;
            }
         }
         
         // start the trim-agents timer
         if (success)
         {
            synchronized (launcher)
            {
               if (launcher.timer == null)
               {
                  launcher.timer = new Timer(appServer + " agent trimming thread.", true);
                  TimerTask tt = new TimerTask()
                  {
                     public void run()
                     {
                        trimAgents(appServer);
                     }
                  };
                  long timeout = appDef.getAutoTrimTimeout() * 1000;
                  launcher.timer.scheduleAtFixedRate(tt, timeout, timeout);
               }
            }
         }
      }
      else
      {
         success = launchWorker(appServer, launcher);
      }
      
      return success;
   }

   /**
    * Terminate the given appserver.
    * 
    * @param    appServer
    *           The name of the appserver to terminate.
    */
   public static void terminate(String appServer)
   {
      AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
      pool.terminateAll();
   }
   
   /**
    * Trim the agents for the specific appserver down to a <code>min_agents</code> count.
    * 
    * @param    appServer
    *           The appserver name.
    */
   public static void trimAgents(String appServer)
   {
      AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
      pool.trimAgents();
   }

   /**
    * Notify that the given appserver has been terminated (there are no more agents for it and
    * the pool is shutting down).
    * 
    * @param    appServer
    *           The name of the appserver being terminated.
    */
   static void terminated(String appServer)
   {
      Launcher launcher = null;
      synchronized (launchers)
      {
         launcher = launchers.remove(appServer);
      }
      
      if (launcher != null)
      {
         synchronized (launcher)
         {
            launcher.timer.cancel();
         }
      }
   
      AppServerDefinition.remove(appServer);
   }

   /**
    * Instantiate a new {@link ProcessBuilder builder} used to launch appserver Agents.  The 
    * command will be built using the following:
    * <ul>
    *    <li>The library path will be set to the folder where the jar containing the 
    *        {@link ClientDriver} class is located.</li>
    *    <li>The classpath will be inherited from the current JVM instance.</li>
    *    <li>The user dir will be set to the directory specified by the <code>user_dir</code>
    *        appserver setting.</li>
    *    <li>The <code>jvm_args</code> appserver setting will specify a list of JVM arguments 
    *        to pass to the P2J Client instance.</li>
    *    <li>If the user dir contains a client.xml file, it will be used.</li>
    *    <li>The target P2J server will be set to this server's secure and insecure ports, 
    *        depending on this server's configuration.</li>
    *    <li>The subject ID needed to be used when authenticated will be passed using the
    *        <code>access:subject:id=&lt;account&gt;</code>argument.</li>
    * </ul>
    * 
    * @param    account
    *           The account used to authenticate the Agent.
    * @param    appServer
    *           The name of the appserver.
    *           
    * @return   A {@link ProcessBuilder builder} to start Agents.
    */
   private static ProcessBuilder builder(String account, String appServer)
   {
      AppServerDefinition appDef = AppServerDefinition.get(appServer);
      String libPath = null;
      try
      {
         File jarFile = new File(ClientDriver.class.getProtectionDomain().getCodeSource()
                                             .getLocation().toURI().getPath());
         // if the source is in a folder and not jar, than this can pose problems
         if (!jarFile.exists())
         {
            // as this is not a valid location, fail
            if (LOG.isErrorEnabled())
            {
               final String msg =
                  "Error starting an Agent for appserver '%s': could not determine the " +
                  "location of the p2j.jar file.";
               throw new IllegalStateException(String.format(msg, appServer));
            }
         }

         libPath = jarFile.getParentFile().getPath();
      }
      catch (URISyntaxException e)
      {
         final String msg =
            "Error starting an Agent for appserver '%s': could not determine the location of the " +
            "p2j.jar file.";
         throw new IllegalStateException(String.format(msg, appServer));
      }

      RuntimeMXBean bean = ManagementFactory.getRuntimeMXBean();
      String classPath = bean.getClassPath();

      String userDir = appDef.getUserDir();
      
      File fUserDir = new File(userDir);
      if (!fUserDir.exists())
      {
         final String msg = "The user dir '%s' specified for appserver '%s' does not exist.";
         throw new IllegalStateException(String.format(msg, userDir, appServer));
      }
      
      ProcessBuilder pb = new ProcessBuilder();
      // set the working dir to be the same as this dir's working dir
      pb.directory(fUserDir);
      // environment is automatically set as a copy of this proc's env
      
      // build the command
      List<String> command = new ArrayList<String>();

      command.add("java");
      // add JVM args; this includes custom environment properties, via -Dprop=value
      String[] jvmArgs = appDef.getJvmArguments();
      for (int i = 0; i < jvmArgs.length; i++)
      {
         command.add(jvmArgs[i]);
      }
      
      // add library and class paths
      command.add(String.format("-Djava.library.path=%s", libPath));
      command.add("-classpath");
      command.add(classPath);
      
      // add main class
      command.add(ClientDriver.class.getName());
      
      // add configuration file
      String clientConfig = String.format("%s%sclient.xml", userDir, File.separator);
      if (new File(clientConfig).exists())
      {
         command.add("client.xml");
      }
      else if (LOG.isWarnEnabled())
      {
         final String msg = "%s does not exist for appserver '%s'";
         LOG.warn(String.format(msg, clientConfig, appServer));
      }
      
      // add target server ports
      BootstrapConfig serverConfig = SessionManager.get().config();
      boolean isSecure = serverConfig.getBoolean("net", "connection",  "secure", false);
      boolean isInsecure = serverConfig.getBoolean("net", "connection", "insecure", true);
      
      if (isInsecure)
      {
         int insecurePort = serverConfig.getInt("net", "server", "port", -1);
         if (insecurePort == -1)
         {
            insecurePort = serverConfig.getInt("net", "server", "insecure_port", -1);
         }
         
         if (insecurePort != -1)
         {
            command.add("net:server:insecure_port=" + insecurePort);
         }
      }
      if (isSecure)
      {
         int securePort = serverConfig.getInt("net", "server", "secure_port", -1);
         
         if (securePort != -1)
         {
            command.add("net:server:secure_port=" + securePort);
         }
      }
      
      // set account
      command.add("access:subject:id=" + account);

      pb.command(command);

      return pb;
   }

   /**
    * Worker for Agent launching.  After starting the Agent via the {@link Launcher#builder},
    * it will wait for {@link #LAUNCH_TIMEOUT} seconds for the Agent to call {@link #notifyLaunch}.
    * <p>
    * After the timeout  or when it receives the launch notification, it will check the 
    * {@link Launcher#success} flag, for the status of the launch.  Note that the {@link Launcher}
    * is shared among processes trying to launch Agents; thus, the launch notification may come
    * from a different Agent than the one currently starting.  This is not a problem, the party
    * starting the Agent doesn't care which agent started, it will only need to know if an Agent
    * was started; other parties waiting for Agent startup will eventually pick up the launch
    * notification, if the Agent was started K.
    * 
    * @param    appServer
    *           The name of the appserver to be launched.
    * @param    launcher
    *           The launcher used to start the agent.
    *           
    * @return   <code>true</code> if the Agent was started successfully.
    */
   private static boolean launchWorker(String appServer, Launcher launcher)
   {
      boolean success = false;
      synchronized (launcher)
      {
         try
         {
            // no pooling is needed here. if an agent needs to be terminated, just send the 
            // TERMINATE command to it
            if (LOG.isInfoEnabled())
            {
               final String msg = "Start command sent for appserver '%s': [%s].";
               StringBuilder sb = new StringBuilder();
               for (String arg : launcher.builder.command())
               {
                  sb.append(arg).append(" ");
               }
               LOG.info(String.format(msg, appServer, sb.toString().trim()));
            }

            launcher.builder.start();
         }
         catch (Throwable t)
         {
            if (LOG.isErrorEnabled())
            {
               final String msg = "Error starting agent for appserver '%s' with command %s.";
               LOG.error(String.format(msg, appServer,  launcher.builder.command().toString()), t);

               return false;
            }
         }
         
         try
         {
            launcher.wait(LAUNCH_TIMEOUT);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
         
         success = (launcher.success != null && launcher.success);

         // reset the value, so is not seen by other threads
         launcher.success = null;
      }
      
      if (LOG.isInfoEnabled())
      {
         final String msgSuccess = "Appserver '%s' was started sucessfully!";
         final String msgFailure = "Appserver '%s' could not be started!";
         LOG.info(String.format(success ? msgSuccess : msgFailure, appServer));
      }
      
      return success;
   }
   
   /**
    * Container for a Agent launching information.
    */
   private static class Launcher
   {
      /** Agent launcher instance. */
      private final ProcessBuilder builder;
      
      /**
       * When set to {@link Boolean#TRUE}, it will indicate the Agent was started successfully;
       * any other value will indicate failure.
       */
      private Boolean success;

      /** Timer for trimming the Agents. */
      private Timer timer;
      
      /**
       * Create a new Agent launcher.
       * 
       * @param    builder
       *           The Agent builder instance.
       */
      private Launcher(ProcessBuilder builder)
      {
         this.builder = builder;
      }
   }
}
